#include <stdio.h>
#include <stdlib.h>

#include "imagen.h"
#include <string.h>

void promedio(const Imagen* img_in, int radio, Imagen* img_out);

int main(int argc, char* argv[])
{
  Imagen imagen_in, imagen_out;
  int radio;
  int res;
  /* interpretacion de comandos de entrada */
  if (argc < 4) {
    printf("Faltan parámetros!\n");
    printf("Uso:\t%s radio entrada.pgm salida.pgm\n\n",argv[0]);
    exit(-1);
  }
  radio = atoi(argv[1]);
  res = leer_imagen(argv[2],&imagen_in);
  if (res != PNM_OK)  {
    printf("Error al leer imagen %s: %d, %s\n",argv[2],res,mensaje_error(res));
    exit(res);
  }
  duplicar_imagen(&imagen_in, &imagen_out);
  promedio(&imagen_in, radio, &imagen_out);
  res = escribir_imagen(&imagen_out, argv[3]);
  if (res != PNM_OK) {
    printf("Error al escribir archivo %s. Codigo %d, mensaje %s\n",argv[3],res,mensaje_error(res));
  }
  destruir_imagen(&imagen_in);
  destruir_imagen(&imagen_out);
  return res;
}

#ifndef MACRO_CLIP
int clip(int i, int min, int max) {
  return  i <= min ? min : (i >= max ? max : i);
}
#else
#define clip(i, min, max) (i)<=(min) ? (min) : (i)>=(max) ? (max) : (i)  
#endif

#if VER==1

void promedio(const Imagen* img_in, int radio, Imagen* img_out) 
{
  int i,j,di,dj,i2,j2;
  const int ancho = img_in->ancho;
  const int alto  = img_in->alto;
  Pixel* pin = img_in->pixels;
  Pixel* pout = img_out->pixels;
  double a;
  printf("VERSION 1: CUENTAS REPETIDAS, calculamos el área en cada ieración\n");
  for (i = 0; i < alto ; ++i) {
    for (j = 0 ; j < ancho; ++j) {
      a = 0.0;
      for (di = -radio; di <= radio ; ++di) {        
        for (dj = -radio ; dj <= radio; ++dj) {
          i2 = clip(i + di, 0, alto-1 );
          j2 = clip(j + dj, 0, ancho-1 );
          a += pin[i2*ancho+j2];
        }
      }
      pout[i*ancho+j] = (int) (a / ((2*radio+1)*(2*radio+1)) + 0.5);
    }
  }
}

#elif VER==2
void promedio(const Imagen* img_in, int radio, Imagen* img_out) 
{
  int i,j,di,dj,i2,j2;
  const int ancho = img_in->ancho;
  const int alto  = img_in->alto;
  Pixel* pin = img_in->pixels;
  Pixel* pout = img_out->pixels;
  printf("VERSION 2: División en cada iteración\n");
  double C  = ((double)((2*radio+1)*(2*radio+1)));
  
  for (i = 0; i < alto ; ++i) {
    for (j = 0 ; j < ancho; ++j) {
      register int a  = 0;
      for (di = -radio; di <= radio ; ++di) {        
        for (dj = -radio ; dj <= radio; ++dj) {
          i2 = clip(i + di, 0, alto-1 );
          j2 = clip(j + dj, 0, ancho-1 );
          a += pin[i2*ancho+j2];
        }
      }
      pout[i*ancho+j] = (int) (a / C + 0.5);
    }
  }
}
#elif VER==3
void promedio(const Imagen* img_in, int radio, Imagen* img_out) 
{
  int i,j,di,dj,i2,j2;
  const int ancho = img_in->ancho;
  const int alto  = img_in->alto;
  Pixel* pin = img_in->pixels;
  Pixel* pout = img_out->pixels;
  double C  = 1.0/((double)((2*radio+1)*(2*radio+1)));
  printf("VERSION 3: CONSTANTE DIVISORIA\n");
  
  for (i = 0; i < alto ; ++i) {
    for (j = 0 ; j < ancho; ++j) {
      register int a = 0;
      for (di = -radio; di <= radio ; ++di) {        
        for (dj = -radio ; dj <= radio; ++dj) {
          i2 = clip(i + di, 0, alto-1 );
          j2 = clip(j + dj, 0, ancho-1 );
          a += pin[i2*ancho+j2];
        }
      }
      pout[i*ancho+j] = (int) (C * a + 0.5);
    }
  }
}
#elif VER==4
void promedio(const Imagen* img_in, int radio, Imagen* img_out) 
{
  int i,j,di,dj,i2,j2;
  const int ancho = img_in->ancho;
  const int alto  = img_in->alto;
  Pixel* pin = img_in->pixels;
  Pixel* pout = img_out->pixels;
  double a;
  double C  = 1.0/((double)((2*radio+1)*(2*radio+1)));
  /* limites entre tramos */
  const int jmax1 =  radio < ancho ? radio: ancho;
  const int jmax2 =  ancho - radio;
  const int imax1 =  radio < alto ? radio: alto;
  const int imax2 =  alto - radio;
  printf("VERSION 4: POR REGIONES\n");
  /* arriba */
  for (i = 0; i < imax1 ; ++i) {
    /* arriba - izq */
    for (j = 0 ; j < jmax1; ++j) {
      a = 0.0;
      for (di = -radio; di <= radio ; ++di) {        
        for (dj = -radio ; dj <= radio; ++dj) {
          i2 = clip(i + di, 0, alto-1 );
          j2 = clip(j + dj, 0, ancho-1 );
          a += pin[i2*ancho+j2];
        }
      }
      pout[i*ancho+j] = (int) (C * a + 0.5);
    }
    /* arriba - centro */
    for ( ; j < jmax2; ++j) {
      a = 0.0;
      for (di = -radio; di <= radio ; ++di) {        
        for (dj = -radio ; dj <= radio; ++dj) {
          i2 = clip(i + di, 0, alto-1 );
          j2 = j + dj;
          a += pin[i2*ancho+j2];
        }
      }
      pout[i*ancho+j] = (int) (C * a + 0.5);
    }
    /* arriba - der */
    for ( ; j < ancho; ++j) {
      a = 0.0;
      for (di = -radio; di <= radio ; ++di) {        
        for (dj = -radio ; dj <= radio; ++dj) {
          i2 = clip(i + di, 0, alto-1 );
          j2 = clip(j + dj, 0, ancho-1 );
          a += pin[i2*ancho+j2];
        }
      }
      pout[i*ancho+j] = (int) (C * a + 0.5);
    }
  }
  /* medio  */
  for (; i < imax2 ; ++i) {
    /* medio - izq */
    for (j = 0 ; j < jmax1; ++j) {
      a = 0.0;
      for (di = -radio; di <= radio ; ++di) {        
        for (dj = -radio ; dj <= radio; ++dj) {
          i2 = i + di;
          j2 = clip(j + dj, 0, ancho-1 );
          a += pin[i2*ancho+j2];
        }
      }
      pout[i*ancho+j] = (int) (C * a + 0.5);
    }
    /* medio - medio */
    for ( ; j < jmax2; ++j) {
      a = 0.0;
      for (di = -radio; di <= radio ; ++di) {        
        for (dj = -radio ; dj <= radio; ++dj) {
          i2 = i + di;
          j2 = j + dj;
          a += pin[i2*ancho+j2];
        }
      }
      pout[i*ancho+j] = (int) (C * a + 0.5);
    }
    /* medio - der */
    for ( ; j < ancho; ++j) {
      a = 0.0;
      for (di = -radio; di <= radio ; ++di) {        
        for (dj = -radio ; dj <= radio; ++dj) {
          i2 = i + di;
          j2 = clip(j + dj, 0, ancho-1 );
          a += pin[i2*ancho+j2];
        }
      }
      pout[i*ancho+j] = (int) (C * a + 0.5);
    }
  }
  /* abajo */
  for (; i < alto ; ++i) {
    /* abajo - izq */
    for (j = 0 ; j < jmax1; ++j) {
      a = 0.0;
      for (di = -radio; di <= radio ; ++di) {        
        for (dj = -radio ; dj <= radio; ++dj) {
          i2 = clip(i + di, 0, alto-1 );
          j2 = clip(j + dj, 0, ancho-1 );
          a += pin[i2*ancho+j2];
        }
      }
      pout[i*ancho+j] = (int) (C * a + 0.5);
    }
    /* abajo - medio */
    for ( ; j < jmax2; ++j) {
      a = 0.0;
      for (di = -radio; di <= radio ; ++di) {        
        for (dj = -radio ; dj <= radio; ++dj) {
          i2 = clip(i + di, 0, alto-1 );
          j2 = j + dj;
          a += pin[i2*ancho+j2];
        }
      }
      pout[i*ancho+j] = (int) (C * a + 0.5);
    }
    /* abajo - der */
    for (; j < ancho; ++j) {
      a = 0.0;
      for (di = -radio; di <= radio ; ++di) {        
        for (dj = -radio ; dj <= radio; ++dj) {
          i2 = clip(i + di, 0, alto-1 );
          j2 = clip(j + dj, 0, ancho-1 );
          a += pin[i2*ancho+j2];
        }
      }
      pout[i*ancho+j] = (int) (C * a + 0.5);
    }
  }
}
#else
void promedio(const Imagen* img_in, int radio, Imagen* img_out)  {
  int i,j,di,dj,i2,j2;
  const int ancho = img_in->ancho;
  const int alto  = img_in->alto;
  Pixel* pin = img_in->pixels;
  Pixel* pout = img_out->pixels;
  Imagen img_aux;
  printf("VERSION 5: EN DOS PASOS\n");
  duplicar_imagen(img_in,&img_aux);
  Pixel* paux = img_aux.pixels;
  Pixel* pac = paux; /*puntero a acumulador*/
  Pixel* ppixout = pout; /*punero a direccón del pixel de salida que escribimos*/

  double C  = 1.0/((double)((2*radio+1)*(2*radio+1)));
  for (i = 0 ; i < alto ; ++i) {
    for (j = 0 ; j < ancho; ++j, pac++) {
      *pac = 0; /*inicializamos el acumulador*/
      for (dj = -radio ; dj <= radio; ++dj) {
        j2 = clip(j + dj, 0, ancho-1 );
        *pac += pin[i*ancho+j2];
      }
    }
  }
  for (i = 0 ; i < alto ; ++i) {
    for (j = 0 ; j < ancho; ++j,ppixout++) {
      register int a = 0;
      for (di = -radio; di <= radio ; ++di) {        
        i2 = clip(i + di, 0, alto-1 );
        a += paux[i2*ancho+j];
      }
      *ppixout = (int) (C * a + 0.5);
    }
  }
  destruir_imagen(&img_aux);
}

#endif
